<?php

/**
 * @file
 * Provide the Automated Testing tab and UI functions.
 */

/**
 * Produce html code for the Automated Testing tab.
 */
function pift_results_project_tab($node, $version = NULL) {
  // Load array of release branches in this projects repo.
  $versions = pift_get_releases($node);

  if (empty($version) && !empty($versions)) {
    $version = current($versions);
  }

  // Retrieve project test results for the current branch/tag from {PIFT_TEST}
  $tests = pift_results_testresult_load($node, $versions);

  // Determine list of releases without tests
  $noresults = $versions;

  if (!empty($tests)) {
    foreach ($tests as $testdata) {
      foreach ($testdata as $test) {
        unset($noresults[$test['version']]);
      }
    }
    // Format test results
    $statusoutput = pift_results_format_testresult_table($tests);
  }
  else {
    $statusoutput = pift_results_empty_text();
  }

  // Okay we have the variable data let's build the page.
  $content = array();
  $content[] = '<p>' . t('This page provides information regarding automated testing status for this project\'s releases.') . '</p>';

  // Add the 'Testing Status' table
  $content[] = '<h3>' . t('Branch Test Results') . '</h3>';
  $content[] =  $statusoutput;

  // If there are any untested branches, provide a form which can be
  // used to initiate an 'on-demand' branch test.
  if (!empty($noresults)) {
    $content[] = '<br /><br />' . drupal_get_form('pift_results_test_branch_form', $node, $noresults);
  }

  // TODO:  (Future development) If test has been queued or sent for too
  // long, provide a form which will allow the user to reset the test (by
  // updating status for tests with id = branch_nid in the pift_test table)

  return implode("\n", $content);
}

/**
 * Prepare a form API array to mark a specific module version for testing
 *
 * @param $form_state
 * @param $project object The project node object
 * @param versions array List of versions
 * @return array
 */
function pift_results_test_branch_form(&$form_state, $project, $versions = array()) {
  $form = array(
    '#prefix' => '<div class="container-inline">',
    '#suffix' => '</div>',
  );

  $description = '<strong>' . t('Note: Automated tests are typically run on ".x" branches.') . '</strong><br />';
  $description .= t('Core examples: 6.x, 7.x, 8.x') . '<br />';
  $description .= t('Contrib examples: 6.x-1.x, 7.x-2.x, 8.x-1.x') . '<br />';
  $description = t('Marking a branch for testing will add it to the list of branch/patch requests sent to the automated testing infrastructure. ');
  $description .= t('Once the tests complete, the results will be passed back to drupal.org and reflected on this page. ');
  $description .= '<br />';

  $form['nid'] = array(
    '#type' => 'value',
    '#value' => $project->nid,
  );

  $form['additional'] = array(
    '#type' => 'fieldset',
    '#title' => t('Test Additional Branches'),
    '#collapsible' => TRUE,
    '#collapsed' => TRUE,
    '#prefix' => '<div>',
    '#suffix' => '</div>',
    '#description' => $description,
  );

  $form['additional']['version'] = array(
    '#type' => 'select',
    '#title' => t('Select a branch/tag to queue for testing'),
    '#options' => $versions,
    '#required' => TRUE,
  );

  $form['additional'][] = array(
    '#type' => 'submit',
    '#value' => t('Queue Test'),
  );

  return $form;
}

/**
 * Validate the submitted pift_results_test_branch_form form
 *
 * @param $form form to validate
 * @param $form_state A keyed array containing the current state of the form.
 */
function pift_results_test_branch_form_validate($form, &$form_state) {
  // Ensure required values have been passed
  if (!$form_state['values']['nid'] || !$form_state['values']['version']) {
    form_set_error('form', t('Error: Unable to determine correct project tag/branch.'));
  }
}

/**
 * Process the submitted pift_results_test_branch_form form
 *
 * @param $form form to submit
 * @param $form_state A keyed array containing the current state of the form.
 */
function pift_results_test_branch_form_submit($form, &$form_state) {
  pift_results_test_add($form_state['values']['nid'], $form_state['values']['version']);
}

/**
 * Queues a particular project branch/tag/release for testing
 *
 * @param $nid The node id for the given project to be tested
 * @param $version A label representing a particular project branch/tag/release
 */
function pift_results_test_add($nid, $version) {

  // Get the repo_id for the project
  $repo_id = db_result(db_query("SELECT repo_id FROM {versioncontrol_project_projects} WHERE nid = %d", $nid));

  // Get the label_id for the passed version
  $label_id = db_result(db_query("SELECT label_id FROM {versioncontrol_labels} WHERE repo_id = %d AND name = '%s'", $repo_id, $version));

  /********************* Version Control Label ID code ********************
  // For use if/when PIFT/PIFR are refactored to use label_id instead of release_nid.
  if ($label_id) {
    db_query('INSERT INTO {pift_test} (type, id, status) VALUES (%d, %d, %d)', PIFT_TYPE_RELEASE, $label_id, PIFT_STATUS_QUEUE);
    drupal_set_message(t("This project branch/tag has been queued for testing."));
  }
  else {
    drupal_set_message("Unable to queue this project branch/tag for testing, due to an error in resolving the corresponding label ID.", 'error');
  }
  */
  /********************* Release Nid code ********************/
  $release_nid = db_result(db_query("SELECT release_nid FROM {versioncontrol_release_labels} WHERE label_id = %d AND project_nid = %d", $label_id, $nid));
  if ($release_nid) {
    db_query('INSERT INTO {pift_test} (type, id, status) VALUES (%d, %d, %d)', PIFT_TYPE_RELEASE, $release_nid, PIFT_STATUS_QUEUE);
    drupal_set_message("This project branch/tag has been queued for testing.");
  }
  else {
    drupal_set_message("This branch does not have an associated release node.  Unable to queue branch for testing.", 'error');
  }
}

/**
 * Retrieves recent testing status for a given project branch/tag/release
 *
 * @param $node The node id for the given project
 * @param $versions An array of project branch/tag/release labels
 *
 * @return $status Array of test results, or FALSE if no results available
 */
function pift_results_testresult_load($node, $versions) {
  // Ensure we've been passed an actual project node
  if ($node->type != 'project_project') {
    return FALSE;
  }
  // Initialize placeholder for test results return value
  $tests = array();

  // Get the repo_id for the project
  $repo_id = db_result(db_query("SELECT repo_id FROM {versioncontrol_project_projects} WHERE nid = %d", $node->nid));

  // Get the label_ids for the passed versions
  $label_ids = array();
  $result = db_query("SELECT label_id, name FROM {versioncontrol_labels}
          WHERE repo_id = %d AND name IN (" . db_placeholders($versions, 'text') . ")",
          array_merge( (array) $repo_id, $versions));
  while ($data = db_fetch_object($result)) {
    $label_ids[$data->name] = $data->label_id;
  }

  /********************* Version Control Label ID code ********************
  // Get the branch test results for this label_id
  $sql = "Select * from {pift_test} where type = 1 and id IN (" . db_placeholders($label_ids, 'int') . ")";
  $result = db_query($sql, $label_ids);
  */
  /********************* Release Nid code ********************/
  // Retrieve corresponding release nids for the requested labels
  $release_nids = array();
  $result = db_query("SELECT release_nid, label_id FROM {versioncontrol_release_labels}
          WHERE project_nid = %d AND label_id IN (" . db_placeholders($label_ids, 'int') . ")",
          array_merge( (array) $node->nid, $label_ids));
  while ($data = db_fetch_object($result)) {
    $release_nids[$data->label_id] = $data->release_nid;
  }

  // Obtain the actual test results
  $sql = "SELECT test_id, status, message, id, last_tested FROM {pift_test}
          WHERE type = 1 AND id IN (" . db_placeholders($release_nids, 'int') . ")
          ORDER BY last_tested DESC";
  $result = db_query($sql, $release_nids);
  while ($data = db_fetch_array($result)) {
    $data['version'] = array_search(array_search($data['id'], $release_nids), $label_ids);
    unset($data['id']);
    $tests[$data['version']][] = $data;
  }
  if (!empty($tests)) {
    // Re-order tests array
    $sorted = array();
    foreach ($versions as $version) {
      if (array_key_exists($version, $tests)) {
        $sorted[] = $tests[$version];
      }
    }
    return $sorted;
  }
  return FALSE;
}

/**
 * Formats the testing results table for output
 *
 * @param $tests An array of test statuses to be rendered
 *
 * @return Themed table containing the testing status for the given branch
 */
function pift_results_format_testresult_table($tests = array()) {
  if (!empty($tests)) {
    $header = array('Version', 'Status', 'Result', 'Last Tested', 'Operations');
    foreach ($tests as $testdata) {
      foreach ($testdata as $test) {
        $rows[] = pift_results_build_testresult_row($test);
      }
    }
  }
  return theme('table', $header, $rows);
}

/**
 * Formats a row of the testing results table for output
 *
 * @param $test An array of test status information to render
 *
 * @return Themed table containing the testing status for the given branch
 */
function pift_results_build_testresult_row($test = array()) {
  if (!empty($test)) {
    $options = array('1' => 'Queued', '2' => 'Sent', '3' => 'Failed', '4' => 'Passed');
    $rowclass = array('1' => 'pift-retest', '2' => 'pift-retest', '3' => 'pift-fail', '4' => 'pift-pass');
    $operations = array();
    // A first-time queued test will not have a test_id, so hide links
    if ($test['test_id'] > 0) {
      $operations = array(
        l('View Test', variable_get('pift_server', 'http://qa.drupal.org/') . 'pifr/test/' . $test['test_id']),
      );
      if (user_access('pift re-test files') && $test['status'] > PIFT_STATUS_SENT) {
        $operations[] = l('Re-test', 'pift/retest/' . $test['test_id']);
      }
      if (user_access('pift re-test files') && $test['status'] == PIFT_STATUS_SENT) {
        $operations[] = l('Remove Test', 'pift/delete/' . $test['test_id']);
      }
    }
    // Hide old results if new test queued/sent
    if ($test['status'] <= PIFT_STATUS_SENT) {
      $test['message'] = t('Waiting for results.');
    }
    // Format 'Last Tested' value
    $test['last_tested'] = $test['last_tested'] > 0 ? format_date($test['last_tested'], 'small') : 'never';
    $row = array(
      'data' => array($test['version'], $options[$test['status']], $test['message'], $test['last_tested'], implode(" | ", $operations)),
      'class' => $rowclass[$test['status']],
    );
    return $row;
  }
  return array();
}

/**
 * Returns the default 'no test results available' text
 */
function pift_results_empty_text() {
  $default = <<<EOT
<p>No Test Results Available.</p>
<p>Please use the 'queue test' form below to trigger a new test.</p>
EOT;
  return $default;
}
